/* Median1.cpp : Median Filter for Speckle Reduction

Mean Filter for Speckle Reduction v. 1.1
Created by:  Nathan Jennings
             nate@jenningsplanet.com
Created on:  05.26.2010
Updated on:  05.30.2010

Computes Median value (middle number of a sorted array of values) of a moving window.  
Part of the Enhancment options.


The majority of this code follows the general logic produced by Ball Aerospace.  The significant contributions
to this code are the statistical filters found below.
 
 * The information in this file is
 * Copyright(c) 2007 Ball Aerospace & Technologies Corporation
 * and is subject to the terms and conditions of the
 * GNU Lesser General Public License Version 2.1
 * The license text is available from   
 * http://www.gnu.org/licenses/lgpl.html
 */

#include "DataAccessor.h"
#include "DataAccessorImpl.h"
#include "DataRequest.h"
#include "DesktopServices.h"
#include "MessageLogResource.h"
#include "ObjectResource.h"
#include "PlugInArgList.h"
#include "PlugInManagerServices.h"
#include "PlugInRegistration.h"
#include "Progress.h"
#include "RasterDataDescriptor.h"
#include "RasterElement.h"
#include "RasterUtilities.h"
#include "SpatialDataView.h"
#include "SpatialDataWindow.h"
#include "switchOnEncoding.h"
#include "median1.h"
#include <limits>
#include <stdlib.h>  //contains functions such as sorting routines
#include <math.h>  //required for power and square root computations (not used in this algorithm)

REGISTER_PLUGIN_BASIC(RADARTutorial, Median1);


int compare (const void * x, const void *y)  //for use with qsort
	      { return (*(double*)x - *(double*)y);  //double beccause values in array are double
	      }
namespace
{
   template<typename T>

   
    
   void median(T* pData, DataAccessor pSrcAcc, int row, int col, int rowSize, int colSize)
    //void median (T* pData, DataAccessor pSrcAcc, int rowSize, int colSize)
   {
      
	   double mean = 0;
	   double median = 0;
	   double sum = 0;
	   int val = 0;
	   int i=0; //counter for array
	   int j=0;
	   int count = 0;
	   
	   //double variance = 0;
	   //double skewness = 0;
	   int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
	   double * a = NULL; 
	   a = new double [kernel*kernel];  //not used yet

	  //future improvements will have the ability to index a specific "kernel" size 
      
	      mean = 0;
		  median = 0;
		  sum = 0;
		  val = 0;
		  count = 0;
		    
		     
		     for(i=row-offset ; i <= (row +offset); i++)  //loop over kernel rows
			 {for(j=col-offset; j <= (col +offset); j++)  //loop over kernel cols
		  
		        {
					pSrcAcc->toPixel(i,j);  //access the specific row,col
					VERIFYNRV(pSrcAcc.isValid());
					
		  
           T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
		   
		   //load array for sorting
		   a[count] = val;
		   count += 1;
        

		  } //end of small kernel window  col       //end of filling array
			 } // end of small kernel window row
        
	
         // Sort array from low to high
  
	    // qsort parameters
	    // array, number of elements to sort, size of data type (i.e. bytes), name of comparator function
	  
	  
	   
	   //qsort part of stdlib
	  qsort(a, kernel*kernel, sizeof(double), compare);



	  // Get middle value

	  median = a[(kernel*kernel-1) / 2];
	 
	  
	  *pData = static_cast<T>(median);

      double* pDoubleData = reinterpret_cast<double*>(pData);
      *pDoubleData = median;
		 
      
   }
};

Median1::Median1()
{
   setDescriptorId("{f248db5d-37ec-4520-b880-d812d6ab7521}"); //this is unique for Enhancement Mean Filter
   setName("Speckle Reduction Median Filter");
   setVersion("Median Filter v1.1");
   setDescription("Calculate and return a Median Filter raster element for a one band. "
      "image (raster element).");
   setCreator("Nathan Jennings");
   setCopyright("Copyright (C) 2008, Ball Aerospace & Technologies Corp.; (C) 2010 JenningsPlanet.com");
   setProductionStatus(false);
   setType("Sample");
   setSubtype("Median");
   setMenuLocation("[RADAR Processing]/Speckle Reduction Filters/Median");
   setAbortSupported(true);
}

Median1::~Median1()
{
}

bool Median1::getInputSpecification(PlugInArgList*& pInArgList)
{
   VERIFY(pInArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pInArgList->addArg<Progress>(Executable::ProgressArg(), NULL, "Progress reporter");
   pInArgList->addArg<RasterElement>(Executable::DataElementArg(), "Perform median filter on this data element");
   return true;
}

bool Median1::getOutputSpecification(PlugInArgList*& pOutArgList)
{
   VERIFY(pOutArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pOutArgList->addArg<RasterElement>("Result", NULL);
   return true;
}

bool Median1::execute(PlugInArgList* pInArgList, PlugInArgList* pOutArgList)
{
    int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;

   //StepResource pStep("Skewness 1", "app", "5EA0CC75-9E0B-4c3d-BA23-6DB7157BBD54");
   StepResource pStep("Median Filter", "app", "f248db5d-37ec-4520-b880-d812d6ab7521");   //this is unique for Median Filter
   if (pInArgList == NULL || pOutArgList == NULL)
   {
      return false;
   }
   Progress* pProgress = pInArgList->getPlugInArgValue<Progress>(Executable::ProgressArg());
   RasterElement* pCube = pInArgList->getPlugInArgValue<RasterElement>(Executable::DataElementArg());
   if (pCube == NULL)
   {
      std::string msg = "A raster cube must be specified.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   RasterDataDescriptor* pDesc = static_cast<RasterDataDescriptor*>(pCube->getDataDescriptor());  //define a Raster Data Descriptor
                                                                                                  // to gain access to image information (bands, rows, columns, data type, etc)
   VERIFY(pDesc != NULL);   //check to see if there is an image
   if (pDesc->getDataType() == INT4SCOMPLEX || pDesc->getDataType() == FLT8COMPLEX)  //check data type
   {
      std::string msg = "Median Filter cannot be performed on complex types.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }

   FactoryResource<DataRequest> pRequest;
   pRequest->setInterleaveFormat(BSQ);
   DataAccessor pSrcAcc = pCube->getDataAccessor(pRequest.release());  // create a data accessor to "get image data"

   //I think this would stay the same


   //This is to create an output image.  In this case it is a float of 8 bytes (i.e. a double precision image)
   ModelResource<RasterElement> pResultCube(RasterUtilities::createRasterElement(pCube->getName() +
      "_Median_Result", pDesc->getRowCount(), pDesc->getColumnCount(), FLT8BYTES));   //pDesc->getDataType() - this uses the same data type 
                                                                                      //as the input image (might be an integer, but user may need a float
   if (pResultCube.get() == NULL)
   {
      std::string msg = "A raster cube could not be created.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   FactoryResource<DataRequest> pResultRequest;
   pResultRequest->setWritable(true);   //be able to write data to the output image

     // Does this go here?

   //DataAccessor access = pElement->getDataAccessor();

   DataAccessor pDestAcc = pResultCube->getDataAccessor(pResultRequest.release());

   pDestAcc->getColumn();
   //pDestAcc->nextColumn(1);
   //pDestAcc->nextColumn();

   pDestAcc->getRow();
   pDestAcc->nextRow(1);
   pDestAcc->nextColumn(1); 

  // Would the for loops from your example go here?



   for (unsigned int row = offset; row < pDesc->getRowCount()-offset; ++row)
   {
      if (pProgress != NULL)
      {
         pProgress->updateProgress("Calculating result", row * 100 / pDesc->getRowCount(), NORMAL);
      }
      if (isAborted())
      {
         std::string msg = getName() + " has been aborted.";
         pStep->finalize(Message::Abort, msg);
         if (pProgress != NULL)
         {
            pProgress->updateProgress(msg, 0, ABORT);
         }
         return false;
      }
      if (!pDestAcc.isValid())
      {
         std::string msg = "Unable to access the cube data.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }
      for (unsigned int col = offset; col < pDesc->getColumnCount()-offset; ++col)
      {
 
	     //can the switch value be changed to not include row, col, and pDestAcc->getColumn????
	     //this line would then run the median function above, correct?

         switchOnEncoding(pDesc->getDataType(), median, pDestAcc->getColumn(), pSrcAcc, row, col, pDesc->getRowCount(), pDesc->getColumnCount());
         pDestAcc->nextColumn();  // get next column of data
      }

      pDestAcc->nextRow();
	  pDestAcc->nextColumn();
   }

   if (!isBatch())
   {
      Service<DesktopServices> pDesktop;

      SpatialDataWindow* pWindow = static_cast<SpatialDataWindow*>(pDesktop->createWindow(pResultCube->getName(),
         SPATIAL_DATA_WINDOW));

      SpatialDataView* pView = (pWindow == NULL) ? NULL : pWindow->getSpatialDataView();
      if (pView == NULL)
      {
         std::string msg = "Unable to create view.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }

      pView->setPrimaryRasterElement(pResultCube.get());
      pView->createLayer(RASTER, pResultCube.get());
   }

   if (pProgress != NULL)
   {
      pProgress->updateProgress("Median Filter is compete.", 100, NORMAL);
   }

   pOutArgList->setPlugInArgValue("Median_Filter_Result", pResultCube.release());

   pStep->finalize();
   return true;
}
