/* Regional1.cpp : Regional Filter for Speckle Reduction

Regional Filter Speckle Reduction v. 1.1
Created by:  Nathan Jennings
             nate@jenningsplanet.com
Created on:  05.31.2010
Updated on:  05.31.2010

Computes a "regional" value within a moving window.  The algorithms takes a windows size (e.g. 3x3) then computes
the mean and variance of "subsets" of this window (2x2 pixel groups).  The subset with the smallest variance
is written to the output image.

The majority of this code follows the general logic produced by Ball Aerospace.  The significant contributions

More information on the Regional Filter can be found here:

Nagao, M. and T. Matsuyama. 1979. Edge preserving smoothing., 
Computer Graphics and Image Processing, 9:394-407.

 * The information in this file is
 * Copyright(c) 2007 Ball Aerospace & Technologies Corporation
 * and is subject to the terms and conditions of the
 * GNU Lesser General Public License Version 2.1
 * The license text is available from   
 * http://www.gnu.org/licenses/lgpl.html
 */

#include "DataAccessor.h"
#include "DataAccessorImpl.h"
#include "DataRequest.h"
#include "DesktopServices.h"
#include "MessageLogResource.h"
#include "ObjectResource.h"
#include "PlugInArgList.h"
#include "PlugInManagerServices.h"
#include "PlugInRegistration.h"
#include "Progress.h"
#include "RasterDataDescriptor.h"
#include "RasterElement.h"
#include "RasterUtilities.h"
#include "SpatialDataView.h"
#include "SpatialDataWindow.h"
#include "switchOnEncoding.h"
#include "Regional1.h"
#include <limits>
#include <stdlib.h>  //contains functions such as sorting routines
#include <math.h>  //required for power and square root computations (not used in this algorithm)

REGISTER_PLUGIN_BASIC(RADARTutorial, Regional1);



namespace
{
   template<typename T>

   
    
   void regional(T* pData, DataAccessor pSrcAcc, int row, int col, int rowSize, int colSize)
    
   {
      
	   double mean = 0;
	   double median = 0;
	   double sum = 0;
	   int val1 = 0;
	   int val2 = 0;
	   int i=0; //counter for array
	   int j=0;
	   int count = 0;
	   
	   
	   int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
	   

	    long CornerRow,CornerCol;      // coords of top left corner of subwindow.
		long WinRow, WinCol;           // current row and column in subwindow.
		double WinSum, WinMean, WinVar; // window statistics (sum, mean, variance)
		double LowVar;                  // lowest variance among subwindows
		double  LowMean;                 // mean of lowest subwindow

	  //future improvements will have the ability to index a specific "kernel" size 
     
	      mean = 0;
		  median = 0;
		  sum = 0;
		  val1 = 0;
		  val2 = 0;
		  count = 0;
		  

		  // initialize variables for lowest subwindow statistics

		LowVar=100000;
		LowMean=0;

					// do each subwindow

		for (CornerRow=(row-offset);CornerRow<=row;CornerRow++) {
			for (CornerCol=(col-offset);CornerCol<=col;CornerCol++) {
				WinSum=0;
				WinMean=0;
				WinVar=0;
				val1 = 0;
				val2 = 0;

				// calculate subwindow statistics:

			// calcualte the mean
				for (WinRow=CornerRow;WinRow<=(CornerRow+offset);WinRow++){ 
					for (WinCol=CornerCol;WinCol<=(CornerCol+offset);WinCol++){
						
						pSrcAcc->toPixel(WinRow,WinCol);
					    VERIFYNRV(pSrcAcc.isValid());
						T val1 = *reinterpret_cast<T*>(pSrcAcc->getColumn());
						WinSum+=val1;
					} //end WinCol
				}  //end WinRow

			    //compute mean	
			 WinMean=WinSum/((offset+1)*(offset+1));

			 // calculate the variance
			 for (WinRow=CornerRow;WinRow<=(CornerRow+offset);WinRow++){
				 for (WinCol=CornerCol;WinCol<=(CornerCol+offset);WinCol++){
                    pSrcAcc->toPixel(WinRow,WinCol);
					VERIFYNRV(pSrcAcc.isValid());
					T val2 = *reinterpret_cast<T*>(pSrcAcc->getColumn());
					WinVar+=pow((val2-WinMean),2);
				 
				 }  //end WinCol
			 }   //end WinRow


			 // test to see if this is the most homogenous subwindow so far

				if (WinVar<LowVar) {
					LowVar=WinVar;
					LowMean=WinMean;
					}  //end if

			}  //end CornerCol
		}   //end CornerRow

		    
	  
	  *pData = static_cast<T>(LowMean);   //write out the most homogeneous value

      double* pDoubleData = reinterpret_cast<double*>(pData);
      *pDoubleData = LowMean;
		 
	  
      
   }
};

Regional1::Regional1()
{
   setDescriptorId("{ddbd1712-5668-4a13-885a-2235d72af9eb}"); //this is unique for Enhancement Regional Filter
   setName("Speckle Reduction Regional Filter");
   setVersion("Regional Filter v1.1");
   setDescription("Calculate and return a Regional Filter raster element for a one band. "
      "image (raster element).");
   setCreator("Nathan Jennings");
   setCopyright("Copyright (C) 2008, Ball Aerospace & Technologies Corp.; (C) 2010 JenningsPlanet.com");
   setProductionStatus(false);
   setType("Sample");
   setSubtype("Regional");
   setMenuLocation("[RADAR Processing]/Speckle Reduction Filters/Regional");
   setAbortSupported(true);
}

Regional1::~Regional1()
{
}

bool Regional1::getInputSpecification(PlugInArgList*& pInArgList)
{
   VERIFY(pInArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pInArgList->addArg<Progress>(Executable::ProgressArg(), NULL, "Progress reporter");
   pInArgList->addArg<RasterElement>(Executable::DataElementArg(), "Perform median filter on this data element");
   return true;
}

bool Regional1::getOutputSpecification(PlugInArgList*& pOutArgList)
{
   VERIFY(pOutArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pOutArgList->addArg<RasterElement>("Result", NULL);
   return true;
}

bool Regional1::execute(PlugInArgList* pInArgList, PlugInArgList* pOutArgList)
{
    int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;

   //StepResource pStep("Skewness 1", "app", "5EA0CC75-9E0B-4c3d-BA23-6DB7157BBD54");
   StepResource pStep("Regional Filter", "app", "ddbd1712-5668-4a13-885a-2235d72af9eb");   //this is unique for Regional Filter
   if (pInArgList == NULL || pOutArgList == NULL)
   {
      return false;
   }
   Progress* pProgress = pInArgList->getPlugInArgValue<Progress>(Executable::ProgressArg());
   RasterElement* pCube = pInArgList->getPlugInArgValue<RasterElement>(Executable::DataElementArg());
   if (pCube == NULL)
   {
      std::string msg = "A raster cube must be specified.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   RasterDataDescriptor* pDesc = static_cast<RasterDataDescriptor*>(pCube->getDataDescriptor());  //define a Raster Data Descriptor
                                                                                                  // to gain access to image information (bands, rows, columns, data type, etc)
   VERIFY(pDesc != NULL);   //check to see if there is an image
   if (pDesc->getDataType() == INT4SCOMPLEX || pDesc->getDataType() == FLT8COMPLEX)  //check data type
   {
      std::string msg = "Regional Filter cannot be performed on complex types.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }

   FactoryResource<DataRequest> pRequest;
   pRequest->setInterleaveFormat(BSQ);
   DataAccessor pSrcAcc = pCube->getDataAccessor(pRequest.release());  // create a data accessor to "get image data"



   //This is to create an output image.  In this case it is a float of 8 bytes (i.e. a double precision image)
   ModelResource<RasterElement> pResultCube(RasterUtilities::createRasterElement(pCube->getName() +
      "_Regional_Result", pDesc->getRowCount(), pDesc->getColumnCount(), FLT8BYTES));   //pDesc->getDataType() - this uses the same data type 
                                                                                      //as the input image (might be an integer, but user may need a float
   if (pResultCube.get() == NULL)
   {
      std::string msg = "A raster cube could not be created.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   FactoryResource<DataRequest> pResultRequest;
   pResultRequest->setWritable(true);   //be able to write data to the output image

     
   //DataAccessor access = pElement->getDataAccessor();

   DataAccessor pDestAcc = pResultCube->getDataAccessor(pResultRequest.release());

   pDestAcc->getColumn();
   //pDestAcc->nextColumn(1);
   //pDestAcc->nextColumn();

   pDestAcc->getRow();
   pDestAcc->nextRow(1);
   pDestAcc->nextColumn(1); 



   for (unsigned int row = offset; row < pDesc->getRowCount()-offset; ++row)
   {
      if (pProgress != NULL)
      {
         pProgress->updateProgress("Calculating result", row * 100 / pDesc->getRowCount(), NORMAL);
      }
      if (isAborted())
      {
         std::string msg = getName() + " has been aborted.";
         pStep->finalize(Message::Abort, msg);
         if (pProgress != NULL)
         {
            pProgress->updateProgress(msg, 0, ABORT);
         }
         return false;
      }
      if (!pDestAcc.isValid())
      {
         std::string msg = "Unable to access the cube data.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }
      for (unsigned int col = offset; col < pDesc->getColumnCount()-offset; ++col)
      {
 

         switchOnEncoding(pDesc->getDataType(), regional, pDestAcc->getColumn(), pSrcAcc, row, col, pDesc->getRowCount(), pDesc->getColumnCount());
         pDestAcc->nextColumn();  // get next column of data
      }

      pDestAcc->nextRow();
	  pDestAcc->nextColumn();
   }

   if (!isBatch())
   {
      Service<DesktopServices> pDesktop;

      SpatialDataWindow* pWindow = static_cast<SpatialDataWindow*>(pDesktop->createWindow(pResultCube->getName(),
         SPATIAL_DATA_WINDOW));

      SpatialDataView* pView = (pWindow == NULL) ? NULL : pWindow->getSpatialDataView();
      if (pView == NULL)
      {
         std::string msg = "Unable to create view.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }

      pView->setPrimaryRasterElement(pResultCube.get());
      pView->createLayer(RASTER, pResultCube.get());
   }

   if (pProgress != NULL)
   {
      pProgress->updateProgress("Regional Filter is compete.", 100, NORMAL);
   }

   pOutArgList->setPlugInArgValue("Regional_Filter_Result", pResultCube.release());

   pStep->finalize();
   return true;
}
