/* Skewness1.cpp 

Skewness RADAR Texture Algorithm v. 1.1
Created by:  Nathan Jennings
             nate@jenningsplanet.com
Created on:  04.15.2010
Updated on:  05.30.2010

Computes Skewness Statistical Measure how "shifted" a distribution curve is.
A large positive skewness indicates a "shift" to the right of the mean (i.e. 
the right side of the distribution has a longer tail than the left side).  
A large negative value indicates a shift to the left of the mean (i.e. the left 
side of the distribution has a longer tail than the right).  The resulting skewness 
values can be positive or negative.

From an image processing "texture" point of view large (positive or negative) skewness 
values indicates a trend in the local neighborhood of brightness values.  A broad area of 
negative values can indicate that the original brightness values tend to be small, but similar.  
A broad area of positive values can indicate that the original brightness values tend to be larger, 
but similar.  The resulting skewness image can help in determining some land cover types or assit to 
"segment" the image into areas with distinct texture patterns.


 * The information in this file is
 * Copyright(c) 2007 Ball Aerospace & Technologies Corporation
 * and is subject to the terms and conditions of the
 * GNU Lesser General Public License Version 2.1
 * The license text is available from   
 * http://www.gnu.org/licenses/lgpl.html
 */

#include "DataAccessor.h"
#include "DataAccessorImpl.h"
#include "DataRequest.h"
#include "DesktopServices.h"
#include "MessageLogResource.h"
#include "ObjectResource.h"
#include "PlugInArgList.h"
#include "PlugInManagerServices.h"
#include "PlugInRegistration.h"
#include "Progress.h"
#include "RasterDataDescriptor.h"
#include "RasterElement.h"
#include "RasterUtilities.h"
#include "SpatialDataView.h"
#include "SpatialDataWindow.h"
#include "StringUtilities.h"

#include "switchOnEncoding.h"
#include "skewness1.h"
#include <limits>
#include <math.h>   //needed for power and square root computations

REGISTER_PLUGIN_BASIC(RADARTutorial, Skewness1);

namespace
{
   template<typename T>
   void Skewness(T* pData, DataAccessor pSrcAcc, int row, int col, int rowSize, int colSize)
   {
      
	   double mean = 0;
	   double variance = 0;
	   double stdev = 0;
	   double skewness = 0;
	   double median = 0;
	   double sum = 0;
	   int val = 0;
	   int i=0; //counter for array
	   int j=0;
	   int count = 0;
	   
	   
	   int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
	   int * a = NULL; 
	   a = new int [kernel*2];  //not used yet

	      mean = 0;
		  sum = 0;
		  //val = 0;
		  count = 0;
		  variance = 0;
		  skewness = 0;
		  stdev = 0;
		  
		     
		     for(i=row-offset ; i <= (row +offset); i++)  //loop over kernel rows
			 {for(j=col-offset; j <= (col +offset); j++)  //loop over kernel cols
		  
		        {  
					pSrcAcc->toPixel(i,j);  //access the specific row,col
					VERIFYNRV(pSrcAcc.isValid());
					count += 1;
                  
           
		  
		   
           T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
		   //a[i] = val;

		   sum += val;   //add up the pixels

        

		  } //end of small kernel window  col       
			 } // end of small kernel window row
        
		  
         mean = sum / count; //compute the mean

	      //re-run the loop to compute the variance and skewness sum
          for(i=row-offset ; i <= (row +offset); i++)
			 {for(j=col-offset; j <= (col +offset); j++)
			   {    pSrcAcc->toPixel(i,j);
					VERIFYNRV(pSrcAcc.isValid());
		            T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
                    variance += pow((val - mean), 2);
					skewness += pow((val - mean), 3);
		       }
		     }
  

             stdev = sqrt(variance);
		     skewness = skewness / ((kernel*kernel-1) *pow(stdev, 3));	 //compute the skewness

			 //a power of 3 can produce both positive and negative results and may be small in magnitude


      *pData = static_cast<T>(skewness);
  
	  //May implement some additional scaling mechanisms here. (e.g. log scale with min-max)
	  double* pDoubleData = reinterpret_cast<double*>(pData);
      *pDoubleData = skewness;


   }
};

Skewness1::Skewness1()
{
   setDescriptorId("{6850d5b0-4581-11df-9879-0800200c9a66}"); //this is unique for Skewness
   setName("RADAR Texture Analysis Skewness");
   setVersion("Skewness v1.0");
   setDescription("Calculate and return a skewness texture raster element for a one band "
      "image (raster element).");
   setCreator("Nathan Jennings");
   setCopyright("Copyright (C) 2008, Ball Aerospace & Technologies Corp.; (C) 2010 JenningsPlanet.com");
   setProductionStatus(false);
   setType("Sample");
   setSubtype("Skewness");
   setMenuLocation("[RADAR Processing]/Texture Analysis/Skewness");
   setAbortSupported(true);
}

Skewness1::~Skewness1()
{
}

bool Skewness1::getInputSpecification(PlugInArgList*& pInArgList)
{
   VERIFY(pInArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pInArgList->addArg<Progress>(Executable::ProgressArg(), NULL, "Progress reporter");
   pInArgList->addArg<RasterElement>(Executable::DataElementArg(), "Perform skewness texture on this data element");
   return true;
}

bool Skewness1::getOutputSpecification(PlugInArgList*& pOutArgList)
{
   VERIFY(pOutArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pOutArgList->addArg<RasterElement>("Result", NULL);
   
   return true;




}

bool Skewness1::execute(PlugInArgList* pInArgList, PlugInArgList* pOutArgList)
{

	int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
   //StepResource pStep("Skewness 1", "app", "5EA0CC75-9E0B-4c3d-BA23-6DB7157BBD54");
   StepResource pStep("Skewness", "app", "916eb9e0-3d38-11df-9879-0800200c9a66");   //this is unique for skewness
   if (pInArgList == NULL || pOutArgList == NULL)
   {
      return false;
   }
   Progress* pProgress = pInArgList->getPlugInArgValue<Progress>(Executable::ProgressArg());
   RasterElement* pCube = pInArgList->getPlugInArgValue<RasterElement>(Executable::DataElementArg());
   if (pCube == NULL)
   {
      std::string msg = "A raster cube must be specified.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   RasterDataDescriptor* pDesc = static_cast<RasterDataDescriptor*>(pCube->getDataDescriptor());
   VERIFY(pDesc != NULL);
   if (pDesc->getDataType() == INT4SCOMPLEX || pDesc->getDataType() == FLT8COMPLEX)
   {
      std::string msg = "Skewness cannot be performed on complex types.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }

   FactoryResource<DataRequest> pRequest;
   pRequest->setInterleaveFormat(BSQ);
   DataAccessor pSrcAcc = pCube->getDataAccessor(pRequest.release());

   ModelResource<RasterElement> pResultCube(RasterUtilities::createRasterElement(pCube->getName() +
      "_Skewness_Result", pDesc->getRowCount(), pDesc->getColumnCount(), FLT8BYTES));   //pDesc->getDataType()
                                                                         //FLT8BYTES required for double precision
                                                                         //data type above
   if (pResultCube.get() == NULL)
   {
      std::string msg = "A raster cube could not be created.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   FactoryResource<DataRequest> pResultRequest;
   pResultRequest->setWritable(true);
   DataAccessor pDestAcc = pResultCube->getDataAccessor(pResultRequest.release());

   pDestAcc->getColumn();
   //pDestAcc->nextColumn(1);
   //pDestAcc->nextColumn();

   pDestAcc->getRow();
   pDestAcc->nextRow(offset);  //offsets a row around the output image (i.e. 0s)
   pDestAcc->nextColumn(offset); //offsets a col around the image (i.e. 0s) 


       //processes image data using offset
   for (unsigned int row = offset; row < pDesc->getRowCount()-offset; ++row)
   {
      if (pProgress != NULL)
      {
         pProgress->updateProgress("Calculating result", row * 100 / pDesc->getRowCount(), NORMAL);
      }
      if (isAborted())
      {
         std::string msg = getName() + " has been aborted.";
         pStep->finalize(Message::Abort, msg);
         if (pProgress != NULL)
         {
            pProgress->updateProgress(msg, 0, ABORT);
         }
         return false;
      }
      if (!pDestAcc.isValid())
      {
         std::string msg = "Unable to access the cube data.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }
      for (unsigned int col = offset; col < pDesc->getColumnCount()-offset; ++col)
      {
         switchOnEncoding(pDesc->getDataType(), Skewness, pDestAcc->getColumn(), pSrcAcc, row, col, pDesc->getRowCount(), pDesc->getColumnCount());
         pDestAcc->nextColumn();
      }

      pDestAcc->nextRow();
	  pDestAcc->nextColumn();
   }

   if (!isBatch())
   {
      Service<DesktopServices> pDesktop;

      SpatialDataWindow* pWindow = static_cast<SpatialDataWindow*>(pDesktop->createWindow(pResultCube->getName(),
         SPATIAL_DATA_WINDOW));

      SpatialDataView* pView = (pWindow == NULL) ? NULL : pWindow->getSpatialDataView();
      if (pView == NULL)
      {
         std::string msg = "Unable to create view.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }

      pView->setPrimaryRasterElement(pResultCube.get());
      pView->createLayer(RASTER, pResultCube.get());
   }

   if (pProgress != NULL)
   {

	  
      pProgress->updateProgress("Skewness is compete.", 100, NORMAL);
   }

   pOutArgList->setPlugInArgValue("Skewness_Result", pResultCube.release());

   pStep->finalize();
   return true;
}
