/* Variance1.cpp 

Variance RADAR Texture Algorithm v. 1.1
Created by:  Nathan Jennings
             nate@jenningsplanet.com
Created on:  04.15.2010
Updated on:  05.30.2010

Computes Variance Statistical Measure how "variable" a distribution curve is.
A large variance indicates that pixels within a local neighborhood are different from one another.  A small variance
indicate that pixels are similar to one another and may have a normal distribution.

The resulting variance image can help in determining some land cover types or assit 
to "segment" the image into areas with distinct texture patterns.

The majority of this code follows the general logic produced by Ball Aerospace.  The significant contributions
to this code are the statistical measures found below.

 * The information in this file is
 * Copyright(c) 2007 Ball Aerospace & Technologies Corporation
 * and is subject to the terms and conditions of the
 * GNU Lesser General Public License Version 2.1
 * The license text is available from   
 * http://www.gnu.org/licenses/lgpl.html
 */

#include "DataAccessor.h"
#include "DataAccessorImpl.h"
#include "DataRequest.h"
#include "DesktopServices.h"
#include "MessageLogResource.h"
#include "ObjectResource.h"
#include "PlugInArgList.h"
#include "PlugInManagerServices.h"
#include "PlugInRegistration.h"
#include "Progress.h"
#include "RasterDataDescriptor.h"
#include "RasterElement.h"
#include "RasterUtilities.h"
#include "SpatialDataView.h"
#include "SpatialDataWindow.h"
#include "switchOnEncoding.h"
#include "variance1.h"
#include <limits>
#include <math.h>    //required for power and square root computations

REGISTER_PLUGIN_BASIC(RADARTutorial, Variance1);

namespace
{
   template<typename T>
   void variance(T* pData, DataAccessor pSrcAcc, int row, int col, int rowSize, int colSize)
   {
      
	   double mean = 0;
	   double variance = 0;
	   double median = 0;
	   double sum = 0;
	   int val = 0;
	   int i=0; //counter for array
	   int j=0;
	   int count = 0;
	   
	  
	   int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;
	   int * a = NULL; 
	   a = new int [kernel*2];  //not used yet

	  //future improvements will have the ability to index a specific "kernel" size 
      
	      mean = 0;
		  sum = 0;
		  count = 0;
		  variance = 0;
		  
		     
		     for(i=row-offset ; i <= (row +offset); i++)
			 {for(j=col-offset; j <= (col +offset); j++)
		  
		        {  
					pSrcAcc->toPixel(i,j);  //access the specific row,col
					VERIFYNRV(pSrcAcc.isValid());
					count += 1;
           
           
		  
		   
           T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
		   

		   sum += val;   //add up the pixels

        

		  } //end of small kernel window  col       
			 } // end of small kernel window row
        
		  
      mean = sum / count; //compute the mean

			//re-run loop for variance calculation
          for(i=row-offset ; i <= (row +offset); i++)
			 {for(j=col-offset; j <= (col +offset); j++)
			   {    pSrcAcc->toPixel(i,j);
					VERIFYNRV(pSrcAcc.isValid());
		            T val = *reinterpret_cast<T*>(pSrcAcc->getColumn());
                   variance += pow((val - mean), 2);
		       }
		     }
  
		     variance = variance / ((kernel*kernel)-1);   //compute the variance

	 *pData = static_cast<T>(variance);

      double* pDoubleData = reinterpret_cast<double*>(pData);
      *pDoubleData = variance;
   }
};

Variance1::Variance1()
{
   setDescriptorId("{a5827f09-ead0-4e54-8f5c-a51d4eb088af}"); //this is unique for Variance
   setName("RADAR Texture Analysis Variance");
   setVersion("Variance v1.1");
   setDescription("Calculate and return a variance texture raster element for a one band "
      "image (raster element).");
   setCreator("Nathan Jennings");
   setCopyright("Copyright (C) 2008, Ball Aerospace & Technologies Corp.; (C) 2010 JenningsPlanet.com");
   setProductionStatus(false);
   setType("Sample");
   setSubtype("Variance");
   setMenuLocation("[RADAR Processing]/Texture Analysis/Variance");
   setAbortSupported(true);
}

Variance1::~Variance1()
{
}

bool Variance1::getInputSpecification(PlugInArgList*& pInArgList)
{
   VERIFY(pInArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pInArgList->addArg<Progress>(Executable::ProgressArg(), NULL, "Progress reporter");
   pInArgList->addArg<RasterElement>(Executable::DataElementArg(), "Perform variance texture on this data element");
   return true;
}

bool Variance1::getOutputSpecification(PlugInArgList*& pOutArgList)
{
   VERIFY(pOutArgList = Service<PlugInManagerServices>()->getPlugInArgList());
   pOutArgList->addArg<RasterElement>("Result", NULL);
   return true;
}

bool Variance1::execute(PlugInArgList* pInArgList, PlugInArgList* pOutArgList)
{

    int kernel = 3;             //future enhancements will provide the ability for the use to choose
	                            //window size via a GUI
       short offset = (kernel-1) / 2;

   //StepResource pStep("Skewness 1", "app", "5EA0CC75-9E0B-4c3d-BA23-6DB7157BBD54");
   StepResource pStep("Variance", "app", "a5827f09-ead0-4e54-8f5c-a51d4eb088af");   //this is unique for variance
   if (pInArgList == NULL || pOutArgList == NULL)
   {
      return false;
   }
   Progress* pProgress = pInArgList->getPlugInArgValue<Progress>(Executable::ProgressArg());
   RasterElement* pCube = pInArgList->getPlugInArgValue<RasterElement>(Executable::DataElementArg());
   if (pCube == NULL)
   {
      std::string msg = "A raster cube must be specified.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   RasterDataDescriptor* pDesc = static_cast<RasterDataDescriptor*>(pCube->getDataDescriptor());
   VERIFY(pDesc != NULL);
   if (pDesc->getDataType() == INT4SCOMPLEX || pDesc->getDataType() == FLT8COMPLEX)
   {
      std::string msg = "Variance cannot be performed on complex types.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }

   FactoryResource<DataRequest> pRequest;
   pRequest->setInterleaveFormat(BSQ);
   DataAccessor pSrcAcc = pCube->getDataAccessor(pRequest.release());

   ModelResource<RasterElement> pResultCube(RasterUtilities::createRasterElement(pCube->getName() +
      "_Variance_Result", pDesc->getRowCount(), pDesc->getColumnCount(), FLT8BYTES));  //pDesc->getDataType()
                                                                         //FLT8BYTES required for double precision data
   if (pResultCube.get() == NULL)
   {
      std::string msg = "A raster cube could not be created.";
      pStep->finalize(Message::Failure, msg);
      if (pProgress != NULL) 
      {
         pProgress->updateProgress(msg, 0, ERRORS);
      }
      return false;
   }
   FactoryResource<DataRequest> pResultRequest;
   pResultRequest->setWritable(true);
   DataAccessor pDestAcc = pResultCube->getDataAccessor(pResultRequest.release());

   pDestAcc->getColumn();
   //pDestAcc->nextColumn(1);
   //pDestAcc->nextColumn();

   pDestAcc->getRow();
   pDestAcc->nextRow(offset);  //offsets a row around the output image (i.e. 0s)
   pDestAcc->nextColumn(offset); //offsets a col around the image (i.e. 0s)

     //processes image data using offset
   for (unsigned int row = offset; row < pDesc->getRowCount()-offset; ++row)
   {
      if (pProgress != NULL)
      {
         pProgress->updateProgress("Calculating result", row * 100 / pDesc->getRowCount(), NORMAL);
      }
      if (isAborted())
      {
         std::string msg = getName() + " has been aborted.";
         pStep->finalize(Message::Abort, msg);
         if (pProgress != NULL)
         {
            pProgress->updateProgress(msg, 0, ABORT);
         }
         return false;
      }
      if (!pDestAcc.isValid())
      {
         std::string msg = "Unable to access the cube data.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }
      for (unsigned int col = offset; col < pDesc->getColumnCount()-offset; ++col)
      {
         switchOnEncoding(pDesc->getDataType(), variance, pDestAcc->getColumn(), pSrcAcc, row, col, pDesc->getRowCount(), pDesc->getColumnCount());
         pDestAcc->nextColumn();
      }

      pDestAcc->nextRow();
	  pDestAcc->nextColumn();
   }

   if (!isBatch())
   {
      Service<DesktopServices> pDesktop;

      SpatialDataWindow* pWindow = static_cast<SpatialDataWindow*>(pDesktop->createWindow(pResultCube->getName(),
         SPATIAL_DATA_WINDOW));

      SpatialDataView* pView = (pWindow == NULL) ? NULL : pWindow->getSpatialDataView();
      if (pView == NULL)
      {
         std::string msg = "Unable to create view.";
         pStep->finalize(Message::Failure, msg);
         if (pProgress != NULL) 
         {
            pProgress->updateProgress(msg, 0, ERRORS);
         }
         return false;
      }

      pView->setPrimaryRasterElement(pResultCube.get());
      pView->createLayer(RASTER, pResultCube.get());
   }

   if (pProgress != NULL)
   {
      pProgress->updateProgress("Variance is compete.", 100, NORMAL);
   }

   pOutArgList->setPlugInArgValue("Variance_Result", pResultCube.release());

   pStep->finalize();
   return true;
}
